#!/usr/bin/perl

use Carp;
use CGI qw/:standard *pre *a *font *span/;
use POSIX;
use Socket;
use IO::Handle;
use URI::Escape;
use Text::Tabs;
require "cqual-common.pl";

sub create_file($);

# some globals
my $windowStarted = 0;           # true iff we've started making a window
my $windowName = "";             # the current window name
my $windowType = "";             # the current window type
my $windowUpper = 0;		 # true iff this window should be up
my $windowLower = 0;		 # true iff this window should be down
my $windowHidden = 1;            # true iff this window has no body
my $windowInFile = 0;            # true iff we're marking up a file
my $didJump = 0;                 # true iff we did a jump
my $inBlobs = 0;                 # true iff we're constructing a window
my $inOverlays = 0;              # true iff we're marking up a file
my $charPos = 1;                 # position in file, if we're marking up a file
my %colormap = ( "pam-color-1" => "olive",
		 "pam-color-2" => "teal",
		 "pam-color-3" => "red",
		 "pam-color-4" => "aqua",
		 "pam-color-5" => "purple",
		 "pam-color-6" => "blue",
		 "pam-color-7" => "green",
		 "pam-color-8" => "maroon",
		 "pam-color-tainted" => "red",
		 "pam-color-untainted" => "green",
		 "pam-color-yyyy" => "yellow",
		 "pam-color-yy" => "teal",
		 "pam-color-nonyear" => "maroon",
		 "pam-color-ww50" => "red",
		 "pam-color-rcsdate" => "green",
		 "pam-color-locked" => "red",
		 "pam-color-unlocked" => "green",
	      );
my $cacheurl = "http://$serverName/cgi-bin/cqual-cgi/cqual-cache.pl";

sub file_header($) {
    my $windowName = shift;

    return
	header,
	start_html(-title=>$windowName,
		   -style=>
		   ("body { font-size: 12px; line-height: 14px; }\n".
		    ".myAnchor { position: relative; }"),
		   -script=>"var loaded = false; var lastHighlight = \"\"",
		   -onLoad=>"loaded = true;",
		   -bgcolor=>"white",
		   -base=>("http://$serverName/cgi-bin/cqual-cgi/"),
		   -target=>"hidden",
		   ),
	"\n",
	start_pre();
}

# give a file and its type, make a url for it
sub file_url($$) {
    my $file = shift;
    my $type = shift;

    return ("session=" . uri_escape($dir) . "&" .
	    "click=$file&" .
	    "file=$file&" .
	    "type=$type");
}

sub finish_window() {
    if ($windowStarted and ($windowType eq "file")) {
	# print any remaining piece of the file
	while (<SOURCE_FILE>) {
	    print FILE;
	}
	close SOURCE_FILE;
    }
    print FILE end_pre(), end_html(), "\n";
    close FILE;
    $windowStarted = 0;
}

# Javascript
sub jscript_load($$$) {
    my $which = shift;
    my $where = shift;
    my $anchor = shift;
    return <<END;
if (top.$which.location.search != "?$where") {
  top.$which.loaded = false;
}
top.$which.location = "$cacheurl?$where$anchor";
END
}

sub jscript_change_bg_color() {
    return <<END;
function change_bg_color(doc, name, color) {
  if (doc.getElementById && doc.getElementById(name) != null)
    doc.getElementById(name).style.backgroundColor = color;
  else if (doc.layers && doc.layers[name] != null)
    doc.layers[name].bgColor = color;
  else if (doc.all && doc.all[name] && doc.all[name].style)
    doc.all[name].style.backgroundColor = color;
}
END
}

sub jscript_highlight($$$) {
    my $which = shift;
    my $anchor = shift;
    my $color = shift;
    return ("function change_color() {\n" .
	    "  if (top.$which.loaded) {\n" .
	    "    change_bg_color(top.$which.document, top.$which.lastHighlight, \"white\");\n" .
	    "    change_bg_color(top.$which.document, \"A$anchor\", \"$color\");\n" .
	    "    top.$which.lastHighlight = \"A$anchor\";\n" .
	    "  }\n" .
	    "  else\n" .
	    "    setTimeout(\"change_color()\", 50);\n" .
	    "}\n" .
	    "change_color();\n");
}

# get parameters
$dir = param("session");
$click = param("click");
$file = param("file");
$type = param("type");

# Set STDOUT to automatically flush
STDOUT->autoflush(1);

# switch to directory for the session and set the alarm timer
setDirectory($dir, $type);
$SIG{ALRM} = \&timeout;

# process the click
if (!(-e "$click.html")) {
    # connect to cqual process
    alarm 5;
    socket(SOCK, PF_UNIX, SOCK_STREAM, 0) || timeout();
    connect(SOCK, sockaddr_un("socket")) || timeout();
    alarm 0;
    SOCK->autoflush(1);
    
    if ($type eq "file") {
	$type = ""; # I have no idea why.  It's probably a PAM bug.
    }
    print SOCK "$click\n$file\n$type\n";
    create_file $click;
    open CACHE, ">>cache-info" or
	internal_error("Can't append to cache-info file");
    print CACHE "$click $windowLower $windowHidden\n";
    close CACHE;
}
else {
    open CACHE, "<cache-info" or internal_error("No cache-info file");
    while (<CACHE>) {
	last if (($windowLower, $windowHidden) = m/$click (.) (.)/);
    }
}
if (!$windowHidden) {
    my $url = $cacheurl . "?" . $ENV{QUERY_STRING};
    print
	header(),
	start_html(-script => ($windowLower ?
			       "top.lower.location=\"$url\"" :
			       "top.upper.location=\"$url\"")),
	end_html(),
	"\n";
}
else {
    cat("$click.html");
}

###########################################################################
#                                                                         #
# Translate PAM into HTML                                                 #
#                                                                         #
###########################################################################

# get the next token from the socket.  Returns undef if there are no
# more tokens
@tokens = ();
sub get_token() {
    my $token;
start:
    do {
	$token = shift @tokens;
    } until !((defined $token) and ($token =~ /^\s*$/));
    if (!(defined $token)) {
	my $line = <SOCK>;
	if (!defined $line) {
	    return undef;
	}
	@tokens = &parse_line('\s+|\(|\)', "delimiters", $line);
	goto start;
    }
    $_ = $token;
    return $token;
}

# push a token back onto the stack
sub unget_token($) {
    my $token = shift;
    unshift @tokens, $token;
}

# read source file until given position is reached, and return the resulting
# string
sub readUntil($) {
    my $pos = shift;
    my $string;
    internal_error("Bad character position $pos (at $charPos)")
	if $charPos > $pos;
    if ($charPos == $pos) {
	return "";
    }
    read SOURCE_FILE, $string, $pos - $charPos
	or internal_error("Bad character position $pos (at $charPos)");
    $charPos = $pos;
#    $string =~ s/\n/<!-- -->\n/;
    $string =~ s/\n/ \n/;
    return expand($string);
}

# convert a pam color to a valid html color string
sub to_html_color($) {
    my $pam_color = shift;
    my $html_color = $colormap{$pam_color};
    $html_color = "blue" if !(defined $html_color);
    return $html_color;
}

# Parse (
sub parse_lparen() {
    my $lparen = get_token();
    internal_error("Got $lparen instead of (") if $lparen ne "(";
}

# Parse )
sub parse_rparen() {
    my $rparen = get_token();
    internal_error("Got $rparen instead of )") if $rparen ne ")";
}

# Parse "<text>"
sub parse_text() {
    my $text = get_token();
    internal_error("Missing <text>") if !(defined $text);
    $text =~ s/\$/\\\$/g;
    $text =~ s/\@/\\\@/g;
    $text =~ s/</\&lt;/g;
    $text =~ s/>/\&gt;/g;
    $text = eval $text; # unescape the string
    return $text;
}

# parse a <highlight>
sub parse_highlight() {
    my $highlight = get_token();
    internal_error("Missing <highlight>") if !(defined $highlight);
    return $highlight;
}

# Parse <pos>
sub parse_pos() {
    my $num = get_token();
    internal_error("Missing <pos>") if !(defined $num);
    return $num;
}

# Parse ("<file>" "<type>")
sub parse_window() {
    parse_lparen();
    my $file = parse_text();
    my $type = parse_text();
    parse_rparen();
    return ($file, $type);
}

# <blob> and <overlay-atom>
sub parse_blob() {
    $windowHidden = 0;
    internal_error("No window created before blob") unless $windowStarted;
    internal_error("Neither in blobs nor overlays")
	unless $inBlobs or $inOverlays;
    my $lparen = get_token();
    internal_error("No end to blob") if !(defined $lparen);
    if ($lparen eq ")") {
	unget_token($lparen);
	return undef;
    }
    my $kind = get_token();
    if ($kind =~ /^text$/) {
	internal_error("Got text in overlays") if $inOverlays;
	my $text = parse_text();
	print FILE expand($text);
    }
    elsif ($kind =~ /^file$/) {
	my ($file, $type) = parse_window();
	internal_error("Markup of $file instead of $windowName")
	    unless $file eq $windowName;
    }
    elsif ($kind =~ /^hyper$/) {
	my $name = parse_text();
	my $link_name = linkify_buffer($name);
	parse_lparen();
	my $start = parse_pos();
	my $end = parse_pos();
	parse_rparen();
	parse_lparen();
	my $highlight = to_html_color(parse_highlight());
	parse_rparen();
	my $text = parse_text();
	if ($inBlobs) {
	    # $text is meaningful
	    print FILE a({-href=>$link_name},
			 font({-color=>$highlight}, $text));
	}
	else {
	    # $text should be empty
	    internal_error("Non-empty $text in overlay") unless $text eq "";
	    print FILE readUntil($start);
	    print FILE
		a({-href=>$link_name},
		  font({-color=>$highlight}, readUntil($end)));
	}
    }
    elsif ($kind =~ /^hyper-anchor$/) {
	my $name = parse_text();
	my $link_name = linkify_buffer($name);
	my $anchor = parse_text();
	parse_lparen();
	my $start = parse_pos();
	my $end = parse_pos();
	parse_rparen();
	parse_lparen();
	my $highlight = to_html_color(parse_highlight());
	parse_rparen();
	my $text = parse_text();
	if ($inBlobs) {
	    # $text is meaningful
	    print FILE a({-href=>$link_name},
			 font({-color=>$highlight}, $text));
	}
	else {
	    # $text should be empty
	    internal_error("Non-empty $text in overlay") unless $text eq "";
	    print FILE readUntil($start);
	    print FILE
		a({-name=>$anchor,
		   -href=>$link_name},
		  span({-id=>"A$anchor",
			-class=>"myAnchor"},
		       font({-color=>$highlight}, readUntil($end))));
	}
    }
    elsif ($kind =~ /^markup$/) {
	parse_lparen();
	my $start = parse_pos();
	my $end = parse_pos();
	parse_rparen();
	parse_lparen();
	my $highlight = to_html_color(parse_highlight());
	parse_rparen();
	my $text = parse_text();
	if ($inBlobs) {
	    # $text is meaningful
	    print FILE font({-color=>$highlight}, $text);
	}
	else {
	    # $text should be empty
	    internal_error("Non-empty $text in overlay") unless $text eq "";
	    print FILE readUntil($start);
	    print FILE font({-color=>$highlight}, readUntil($end));
	}
    }
    elsif ($kind =~ /^anchor$/) {
	internal_error("Anchor in blobs") if $inBlobs;
	my $name = parse_text();
	parse_lparen();
	my $start = parse_pos();
	my $end = parse_pos();
	parse_rparen();
	print FILE readUntil($start);
	print FILE
	    a({-name=>$name}, 
	      span({-id=>"A$name",
		    -class=>"myAnchor"},
		   readUntil($end)));
    }

    parse_rparen();
    return 1;
}

# convert an internal link name into a proper URL
sub linkify_buffer($) {
    my $internal_name = shift;
#    return "click=" . uri_escape($internal_name);
    return ("cqual-client.pl?" .
	    "session=" . uri_escape($dir) . "&" .
	    "click=" . uri_escape($internal_name) . "&" .
	    "file=" . uri_escape($windowName) . "&" .
	    "type=$windowType");
}

# Translate PAM from SOCK into HTML stored in $filename.html
sub create_file($) {
    use Text::ParseWords;

    my $filename = shift;

    # parse commands
    while (1) {
	my $token = get_token();
	last if !(defined $token);
	internal_error("Got $token instead of (") if $token ne "(";
	my $command = get_token();
	internal_error("End of file") if !(defined $command);
	if ($command =~ /pam-overlays/) {
	    $inOverlays = 1;
	    while (defined parse_blob()) {
	    }
	    $inOverlays = 0;
	}
	elsif ($command =~ /pam-blobs/) {
	    $inBlobs = 1;
	    while (defined parse_blob()) {
	    }
	    $inBlobs = 0;
	}
	elsif ($command =~ /pam-close-window/) {
	    # do nothing
	    parse_window();
	}
	elsif ($command =~ /pam-change-window/) {
	    # Initialize the window
	    internal_error("Window already started") if $windowStarted;
	    ($windowName, $windowType) = parse_window();
	    $windowStarted = 1;
	    if ($windowType eq "file") {
		if (!(-e "$windowName.html")) {
		    $windowInFile = 1;
		    open SOURCE_FILE, "<$windowName"
			or internal_error("Unable to open $windowName: $!");
		    $charPos = 1;
		    open FILE, ">$windowName.html"
			or internal_error("Can't write $windowName.html: $!");
		}
		else {
		    open FILE, ">$click.html"
			or internal_error("Can't write $click.html: $!");
		}
	    }
	    else {
		open FILE, ">$click.html"
		    or internal_error("Can't write $click.html: $!");
	    }
	    print FILE file_header($windowName);
	}
	elsif ($command =~ /pam-clear-window/) {
	    # do nothing
	    parse_window();
	}
	elsif ($command =~ /pam-set-window-up/) {
	    internal_error("window started before pam-set-window-up")
		if $windowStarted;
	    my ($file, $type) = parse_window();
	    $windowUpper = 1;
	}
	elsif ($command =~ /pam-set-window-down/) {
	    internal_error("window started before pam-set-window-down")
		if $windowStarted;
	    my ($file, $type) = parse_window();
	    $windowLower = 1;
	}
	elsif ($command =~ /pam-split-window/) {
	    my ($file, $type) = parse_window();
	}
	elsif ($command =~ /pam-set-access/) {
	}
	elsif ($command =~ /pam-goto-char/) {
	    internal_error("Got pam-goto-char");
	}
	elsif ($command =~ /pam-goto-anchor/) {
	    my ($file, $type) = parse_window();
	    my $anchor = parse_text();
	    my $where = file_url($file, $type);
	    my $which = $windowLower ? "lower" : "upper";
	    if ($windowInFile) {
		# we were marking up a file; now switch to creating
		# the command buffer
		finish_window();
		$windowInFile = 0;
		$windowHidden = 1;
		internal_error("$click.html already exists")
		    if (-e "$click.html");
		open FILE, ">$click.html";
		print FILE file_header($windowName);
	    }
	    print FILE
		"\n",
		script((jscript_load($which, $where, "#$anchor") .
			jscript_change_bg_color() .
			jscript_highlight($which, $anchor, "silver")));
	    $didJump = 1;
	}
	elsif ($command =~ /pam-debug/) {
	}
	elsif ($command =~ /pam-message/) {
	    my $text = parse_text();
	    internal_error($text) if !$windowStarted;
	}
	elsif ($command =~ /pam-default-path/) {
	}
	elsif ($command =~ /pam-waiting/) {
	}
	else {
	    internal_error("Bad PAM command");
	}
	parse_rparen();
    }
    if ($windowType eq "file") {
	if ($windowInFile) {
	    # we were marking up a file; finish and switch to creating the
	    # command buffer
	    finish_window();
	    $windowInFile = 0;
	    $windowHidden = 1;
	    internal_error("$click.html already exists") if (-e "$click.html");
	    open FILE, ">$click.html";
	}
	if (!$didJump) {
	    # we opened a file and didn't jump, so add some script
	    # switch to that file
	    my $where = file_url($windowName, $windowType);
	    my $which = $windowLower ? "lower" : "upper";
	    print FILE
		file_header($windowName),
		"\n",
		script(jscript_load($which, $where, ""));
	}
    }
    finish_window();
}
