#!/usr/bin/perl

use Carp;

$serverName = "lagaffe.cs.berkeley.edu";

sub error_page($$$) {
    my $error = shift;
    my $message = shift;
    my $solution = shift;

    print
	header(),
	start_html("Error"),
	script("top.document.clear();\n" .
	       "top.document.write(\'" .
	       "<html><head><title>Error</title></head><body>" .
	       h1("$error") .
	       p() .
	       "$message" .
	       p() .
	       ($solution =~ /e-mail/ ?
		"If the problem persists, please e-mail " .
		a({href=>"mailto:jfoster\@cs.berkeley.edu"},
		  "jfoster\@cs.berkeley.edu") :
		$solution =~ /resubmit/ ?
		"Please resubmit your program" :
		"") .
	       "</body></html>" .
	       "\');\n" .
	       "top.document.close();\n"
	       ),
        end_html();}

# html page to display on server timeout
sub timeout() {
    error_page("Error", "Server timeout.", "resubmit");
    confess "timeout";
}

# html page to display for bad form data
sub bad_request() {
    error_page("Error", "Malformed request.", "e-mail");
    confess "bad request";
}

sub internal_error($) {
    my $message = shift;
    error_page("Internal Error", $message, "e-mail");
    confess "internal error";
}

# make sure we're going to the right place
sub setDirectory($$) {
    my $dir = shift;
    my $type = shift;

    if (!($dir =~ /^\/tmp\/file[a-zA-Z0-9]{6}$/) or
	(($type ne "file") and ($type ne "buffer"))) {
	bad_request();
    }
    chdir $dir || die "chdir $dir failed";
}

# echo file to STDOUT
sub cat($) {
    my $filename = shift;
    open CAT_FILE, $filename or internal_error("Can't find cached file.");
    while (<CAT_FILE>) {
	print;
    }
    close CAT_FILE;
}

1;
