#!/usr/bin/perl

use CGI qw/:standard/;
use IO::Socket;
use Expect;

$timeout = 600;        # maximum number of seconds between events

sub logmsg { 
    $logfile = "/var/log/httpd/cqual_log";
    open(LOG, ">>$logfile") or die "Can't write $logfile: $!";
    print LOG "$0 $$: @_ at ", scalar localtime, "\n";
    close LOG;
}

sub panic {
    unshift @_, "panic:";
    logmsg(@_);
    die;
}

sub daemonize {
    open STDIN, "/dev/null" or die "Can't read /dev/null: $!";
    open STDOUT, ">/dev/null" or die "can't write /dev/null: $!";
    setsid or die "Can't start a new session: $!";
    open STDERR, ">&STDOUT" or die "Can't dup stdout: $!";
}

# Only let the server live for a fixed time
$SIG{ALRM} = sub { panic "server lifetime exceeded\n"; };
alarm $timeout;

chomp($cwd = `pwd`);

$name = "socket";
$uaddr = sockaddr_un($name);
$proto = getprotobyname("tcp");

socket(Server, PF_UNIX, SOCK_STREAM, 0) || die "socket: $!";
unlink($name);
bind(Server, $uaddr) || die "bind: $!";
listen(Server, SOMAXCONN) || die "listen: $!";

daemonize();

logmsg "server started on $cwd/$name";

my $waitedpid = 0;   # we'll only launch one child process

sub REAPER {
    $waitedpid = wait;
    $SIG{CHLD} = \&REAPER;
    logmsg "reaped $waitedpid" . ($? ? " with exit $?" : '');
}

$SIG{CHLD} = \&REAPER;

# Now wait for connections and act like a transparent pipe
$exp = undef;
while ($waitedpid == 0) {
    accept(Client, Server);
    next if $waitedpid;

    logmsg "connection on $cwd/$name";
#    alarm 0; # don't time out during this request
    Client->autoflush(1);

    if (!(defined $exp)) {
	# If this is the initial connection, start the sub-process
	logmsg "spawing cqual";
	$exp =
	    Expect->spawn("/var/www/cgi-bin/cqual-cgi/cqual -fpam-mode -fpam-html @ARGV 2>&1");
	sleep 1;
	if (!(defined $exp)) {
	    panic "spawn failed";
	}
	print Client "Analysis running, please wait...", br(), br(), "\n";
#	while (($key,$value) = each %ENV) {
#	    logmsg "$key=$value\n";
#	}
    }
    else {
	# Otherwise, send the click information to the already-running cqual
	my $click = <Client>;  # these include the terminating newline
	my $file = <Client>;
	my $type = <Client>;
	if ($click ne "home\n") {
	    print $exp "click\n$click$file$type";
	    for ($i = 0; $i < 4; $i++) {
		<$exp>; # eat up the echo
	    }
	}
    }

    # Finally, pipe cqual info to server
    while (<$exp>) {
	print Client "$_";
	logmsg "Sent $_";
	last if /\/\/\// or /(pam-waiting)/;
    }
    close Client;

    alarm $timeout; # reset the alarm counter
}
unlink($name);
