#!/usr/bin/perl

use CGI qw/:standard *script/;
use URI::Escape;
use POSIX;
use Socket;
use IO::Handle;

#!/usr/bin/perl

use Carp;
use CGI qw/:standard *pre *a *font *span/;
use POSIX;
use Socket;
use IO::Handle;
use URI::Escape;
use Text::Tabs;
require "cqual-common.pl";

sub make_lattice_config($);

STDOUT->autoflush(1); # make sure output gets flushed right away

# Make a directory to hold our files and switch to it
while (1) {
    $dir = tmpnam();
    last if mkdir($dir, 0700);
}
chdir "$dir" || die "chdir $dir failed";

my $documentRoot = $ENV{'DOCUMENT_ROOT'};

#open OUT, ">env";
#while (($key, $value) = each %ENV) {
#    print OUT "$key=$value\n";
#}
#close OUT;

# load in the file
my $filename = param("filename");
if ($filename eq "none") {
    $filename = param("file");
    $filename =~ s/\\/-/g;
    $fh = upload("file");
    if ($fh) {
	# store the uploaded file
	open(OUT, ">$filename");
	while (<$fh>) {
	    print OUT;
	}
	close(OUT);
    }
}
else {
    !system("cp $documentRoot/cqual-html/$filename .") or
	die "cp $documentRoot/cqual-html/$filename . failed";
}

if ($filename) {
    print
	header,
	start_html(-title=>"cqual running...",
		   -bgcolor=>'white');

    # make a lattice configuration file
    make_lattice_config("lattice");

    defined(my $pid = fork) or die "Can't fork: $!";
    if ($pid) {
	# parent
	sleep 1; # wait for socket to be created
	socket(SOCK, PF_UNIX, SOCK_STREAM, 0) || die "socket: $!";
	connect(SOCK, sockaddr_un("socket"));
	SOCK->autoflush(1);

	# display analysis progress
	while (<SOCK>) {
	    last if /\/\/\//;
	    print "$_", br(), "\n";
	}

	if (/\/\/\//) {
	    $initialURL = ("http://$serverName/cgi-bin/cqual-cgi/cqual-client.pl?" .
			   "session=" . uri_escape($dir) . "&" .
			   "click=home&" .
			   "file=&" .
			   "type=buffer");
	    $frameURL = ("http://$serverName/cgi-bin/cqual-cgi/cqual-cache.pl?" .
			 "session=" . uri_escape($dir) . "&" .
			 "click=FRAME&" .
			 "file=&" .
			 "type=buffer");
	    open FRAME, ">FRAME.html" or die "Can't open FRAME.html: $!";
	    print FRAME
		header(),
		"<html><head><title>Analysis Results</title></head>\n",
		"<frameset rows=\"75%,25%,0\">\n",
		"<frame src=\"$initialURL\" name=\"upper\">\n",
		"<frame src=\"about:blank\" name=\"lower\">\n",
		"<frame src=\"about:blank\" name=\"hidden\" noresize>\n",
		"</frameset>\n",
		"</html>\n";
	    close FRAME;
	    print
		"Analysis finished",
		br(),
		"\n",
		script("window.location.href='$frameURL'"),
		"\n",
		p(),
		"If you have javascript enabled, you should have been ",
		"redirected to the analysis results.  If not, ",
		a({href=>$frameURL}, "click here"),
		".\n",
		end_html();
	}
	else {
	    print
		br(),
		"Analysis failed",
		br(),
		"Please correct the above errors and resubmit your program\n",
		end_html();
	}
    }
    else {
	# child
	# start up cqual server process
	my $option_fs =
	    ((param("fs") =~ /on/i) or (param("locked") =~ /on/i))
		? "-fflow-sensitive" : "";
	my $option_preserve =
	    (param("preserve") =~ /on/i) ? "-fcasts-preserve" : "";
	exec("/var/www/cgi-bin/cqual-cgi/cqual-server.pl -config lattice $option_fs $option_preserve $filename");
	die "exec failed";
    }
}
else {
    my $error = cgi_error();
    print
	header(-status=>$error),
	start_html("Error"),
	h1("Error"),
	strong($error),
	p(),
	"Unable to process request.",
	p(),
	"If the problem persists, please e-mail ",
	a({href=>"mailto:jfoster\@cs.berkeley.edu"},
	  "jfoster\@cs.berkeley.edu"),
	end_html();
}

# make the lattice configuration file, based on the input
sub make_lattice_config($) {
    open LATTICE, ">lattice" or die "Can't open lattice: $!";
    print LATTICE "/* Generated lattice file */\n";

    if (param("const") =~ /on/i) {
	print LATTICE <<END;
partial order {
  const [level = ref, sign = pos]
  \$nonconst [level = ref, sign = neg]

  \$nonconst < const
}

END
    }

    if (param("tainted") =~ /on/i) {
	print LATTICE <<END;
partial order {
  \$untainted [level = value, color = "pam-color-untainted", sign = neg]
  \$tainted [level = value, color = "pam-color-tainted", sign = pos]

  \$untainted < \$tainted
}

END
    }

    if (param("locked" =~ /on/i)) {
	print LATTICE <<END;
partial order [flow-sensitive] {
  \$locked [level = value, color = "pam-color-locked", sign = eq]
  \$unlocked [level = value, color = "pam-color-unlocked", sign = eq]
}

END
    }
    close LATTICE;
}
