# This makefile is meant to be included by another.  It gives you a
# target for running each flow-insensitive cqual regression test and
# stopping at the first one that fails.

# Todo: It would be nice to have this file set up for the other
# directories and other kinds of tests other than flow insensitive.

# Daniel S. Wilkerson dsw@cs.berkeley.edu

ifdef CQUAL_REGR
  $(error CQUAL_REGR should not already be defined.)
endif
ifdef OUT_CQUAL
  $(error OUT_CQUAL should not already be defined.)
endif

ifndef CQUAL_DIR
  $(error You must define CQUAL_DIR)
endif

# dsw: Arg, this doesn't work because even though I define the
# variable, it has to have something in it for ifndef to decide that
# it is in fact defined.
# ifndef TEST_TOCLEAN_DIR
#   $(error TEST_TOCLEAN_DIR should already be defined as a list of dirs that will \
#           have rm -rf run on them by some clean target.)
# endif

CQUAL_REGR :=

# This is supposed to exit with 33, since it is a type signature clash
# for C, but just function overloading for C++.
#CQUAL_REGR += multitype.c

# Forgot why this doesn't work.
#CQUAL_REGR += anon-enum-param.c

# Passing tests
CQUAL_REGR += enum4.c
CQUAL_REGR += array.c
CQUAL_REGR += assign-fancy.c
CQUAL_REGR += assign.c
CQUAL_REGR += backflow.c
CQUAL_REGR += builtin.c
CQUAL_REGR += cast-func.c
CQUAL_REGR += cast-preserve-confusion1.c
CQUAL_REGR += cast-preserve-confusion2.c
CQUAL_REGR += cast-void.c
CQUAL_REGR += cast.c
CQUAL_REGR += cast2.c
CQUAL_REGR += cast3.c
CQUAL_REGR += cast4.c
CQUAL_REGR += castlist.c
CQUAL_REGR += change.c
CQUAL_REGR += char.c
CQUAL_REGR += comma-strange.c
CQUAL_REGR += compoundexpr.c
CQUAL_REGR += conditional.c
CQUAL_REGR += conflict.c
CQUAL_REGR += const-subtyping.c
CQUAL_REGR += const.c
CQUAL_REGR += constfield.c
CQUAL_REGR += constunion.c
CQUAL_REGR += diff-decl1.c
CQUAL_REGR += diff-decl2.c
CQUAL_REGR += double-poly.c
CQUAL_REGR += double.c
CQUAL_REGR += emptyargs.c
CQUAL_REGR += enum.c
CQUAL_REGR += enum2.c
CQUAL_REGR += enum3.c
CQUAL_REGR += enum5.c
CQUAL_REGR += err1.c
CQUAL_REGR += error.c
CQUAL_REGR += error2.c
CQUAL_REGR += error3.c
CQUAL_REGR += error4.c
CQUAL_REGR += error5.c
CQUAL_REGR += field.c
CQUAL_REGR += file1.c
CQUAL_REGR += file2.c
CQUAL_REGR += fn-qual.c
CQUAL_REGR += for.c
CQUAL_REGR += funcall.c
CQUAL_REGR += funptr-compat.c
CQUAL_REGR += gt.c
CQUAL_REGR += increment.c
CQUAL_REGR += initarray.c
CQUAL_REGR += initfield.c
CQUAL_REGR += initlist.c
CQUAL_REGR += initmultidim.c
CQUAL_REGR += initstring.c
CQUAL_REGR += initstruct.c
CQUAL_REGR += initunion.c
CQUAL_REGR += initunion2.c
CQUAL_REGR += initunion3.c
CQUAL_REGR += list.c
CQUAL_REGR += multiassign.c
CQUAL_REGR += multidef.c
CQUAL_REGR += multidef2.c
CQUAL_REGR += multidim.c
CQUAL_REGR += multierr.c
CQUAL_REGR += nomain.c
CQUAL_REGR += null.c
CQUAL_REGR += oldstyle-funaddr.c
CQUAL_REGR += oldstyle-newstyle1.c
CQUAL_REGR += oldstyle-newstyle2.c
CQUAL_REGR += oldstyle-with-no-proto.c
CQUAL_REGR += oldstyle-with-proto.c
CQUAL_REGR += oldstyle.c
CQUAL_REGR += oldstyle2.c
CQUAL_REGR += op-lhs.c
CQUAL_REGR += op.c
CQUAL_REGR += pad.c
CQUAL_REGR += plus.c
CQUAL_REGR += polydecl.c
CQUAL_REGR += polydecl2.c
CQUAL_REGR += polydecl3.c
CQUAL_REGR += prelude.c
CQUAL_REGR += quals.c
CQUAL_REGR += return.c
CQUAL_REGR += simple.c
CQUAL_REGR += simple2.c
CQUAL_REGR += simple3.c
CQUAL_REGR += strict-const.c
CQUAL_REGR += string.c
CQUAL_REGR += struct-cast.c
CQUAL_REGR += struct-nouse.c
CQUAL_REGR += struct.c
CQUAL_REGR += structdef1.c
CQUAL_REGR += structdef2.c
CQUAL_REGR += structinit.c
CQUAL_REGR += structtype.c
CQUAL_REGR += taint-call.c
CQUAL_REGR += taint.c
CQUAL_REGR += trans-po.c
CQUAL_REGR += transparent-union.c
CQUAL_REGR += tunion.c
CQUAL_REGR += undefined-varargs.c
CQUAL_REGR += undefined.c
CQUAL_REGR += undefined2.c
CQUAL_REGR += user.c
CQUAL_REGR += user2.c
CQUAL_REGR += valarray.c
CQUAL_REGR += valstruct.c
CQUAL_REGR += valunion.c
CQUAL_REGR += varargs.c
CQUAL_REGR += varargs2.c
CQUAL_REGR += varargs3.c
CQUAL_REGR += varargs4.c
CQUAL_REGR += varargs5.c
CQUAL_REGR += varargs6.c
CQUAL_REGR += voidconditional.c
CQUAL_REGR += voidpointer.c
CQUAL_REGR += voidpointer2.c
CQUAL_REGR += volatile.c
CQUAL_REGR += cast-only-qual.c

# linux ****

CQUAL_REGR_LINUX :=

CQUAL_REGR_LINUX += rtc.i
CQUAL_REGR_LINUX += signal-i386.i
CQUAL_REGR_LINUX += signal.i

# poly ****

CQUAL_REGR_POLY :=

CQUAL_REGR_POLY += copy_from_user-struct-field.c
CQUAL_REGR_POLY += deeppoly-intptrcast.c
CQUAL_REGR_POLY += deeppoly.c
CQUAL_REGR_POLY += fnptr-intptrcast.c
CQUAL_REGR_POLY += fwdbak.c
CQUAL_REGR_POLY += id.c
CQUAL_REGR_POLY += joydev.c
CQUAL_REGR_POLY += multisummary.c
CQUAL_REGR_POLY += op_deref.c
CQUAL_REGR_POLY += op_deref2.c
CQUAL_REGR_POLY += user-equals-kernel.c

# where to put the test output
OUT_CQUAL := test2-cqual
$(OUT_CQUAL):; mkdir $@
TEST_TOCLEAN_DIR += $(OUT_CQUAL)
.PRECIOUS: $(OUT_CQUAL)/%
$(OUT_CQUAL)/%:
	mkdir -p $(OUT_CQUAL)/$(*D)
	ln $(CQUAL_DIR)/tests/$* $(OUT_CQUAL)/$(*D)

.PHONY: cqual-check-regression-basic
cqual-check-regression-basic: $(addprefix cqual-check-regression-basic/,$(CQUAL_REGR))
.PHONY: cqual-check-regression-basic/%
cqual-check-regression-basic/%: $(OUT_CQUAL) $(OUT_CQUAL)/%
# You must cd down into the test directory like this because the
# output from cqual contains the filename
	cd $<; ../$(CQUAL_DIR)/src/cqual -config ../$(CQUAL_DIR)/config/lattice \
          -fprint-results $* > $*.out 2>&1
	diff -u $(CQUAL_DIR)/tests/ok/$*.out $</$*.out
	rm -f $</$*

.PHONY: cqual-check-regression-linux
cqual-check-regression-linux: $(addprefix cqual-check-regression-linux/linux/,$(CQUAL_REGR_LINUX))
.PHONY: cqual-check-regression-linux/%
cqual-check-regression-linux/%: $(OUT_CQUAL) $(OUT_CQUAL)/%
# You must cd down into the test directory like this because the
# output from cqual contains the filename
	cd $<; ../$(CQUAL_DIR)/src/cqual -config ../$(CQUAL_DIR)/config/kernel_lattice \
          -fprint-results $* > $*.out 2>&1
	diff -u $(CQUAL_DIR)/tests/ok/$*.out $</$*.out
	rm -f $</$*

.PHONY: cqual-check-regression-poly
cqual-check-regression-poly: $(addprefix cqual-check-regression-poly/poly/,$(CQUAL_REGR_POLY))
.PHONY: cqual-check-regression-poly/%
cqual-check-regression-poly/%: $(OUT_CQUAL) $(OUT_CQUAL)/%
# You must cd down into the test directory like this because the
# output from cqual contains the filename
	cd $<; ../$(CQUAL_DIR)/src/cqualpoly -fpoly -fcasts-preserve \
          -config ../$(CQUAL_DIR)/config/lattice -fprint-results $* > $*.out 2>&1
	diff -u $(CQUAL_DIR)/tests/ok/$*.out $</$*.out
	rm -f $</$*

.PHONY: cqual-check-regression
cqual-check-regression: cqual-check-regression-basic
cqual-check-regression: cqual-check-regression-linux
cqual-check-regression: cqual-check-regression-poly
