#include "hash-serialize.h"
#include "dd_list.h"

void s18n_hash_table_iterator(s18n_type *htt, void *hashtable, void *iterator, void *arg)
{
  hash_table_scanner **hts = iterator;
  *hts = malloc(sizeof(hash_table_scanner));
  hash_table_scan(hashtable, *hts);
}

int s18n_hash_table_next(s18n_type *htt, void *hashtable, void *iterator, 
			 void *key, void *value, void *arg)
{
  hash_table_kind *htk = (hash_table_kind *)
    ((s18n_hashtable_info *)htt->private_info)->private_info;
  hash_table_scanner **hts = iterator;

  if (*hts == NULL)
    return 0;

  while (hash_table_next(*hts, key, value))
    if (htk->hp == NULL || htk->hp->f == NULL ||
	htk->hp->f(*(void **)key, *(void **)value, htk->hp->arg1, htk->hp->arg2) == 0)
      return 1;

  free(*hts);
  *hts = NULL;
  return 0;
}

void *s18n_hash_table_constructor(s18n_type *htt, int nelts, void *arg)
{
  hash_table_kind *htk = (hash_table_kind *)
    ((s18n_hashtable_info *)htt->private_info)->private_info;
  return make_hash_table(*htk->rhash, htk->need_vals, nelts, htk->hash, htk->cmp);
}

void s18n_hash_table_insert(s18n_type *htt, void *hashtable, void *key, void *value, void *arg)
{
  hash_table_insert(hashtable, key, value);
}

s18n_hashtable_ops hash_ops = {
  .iterator = s18n_hash_table_iterator,
  .next = s18n_hash_table_next,
  .constructor = s18n_hash_table_constructor,
  .insert = s18n_hash_table_insert
};

/* This predicate function is a convenience for hashset */
int hashset_pred_adapter(void *key, void *value, void *arg1, void *arg2)
{
  dd_cmp_fn f = (dd_cmp_fn)arg1;
  return f != NULL && f(key, arg2);
    
}

