/* This file is part of cqual.
   Copyright (C) 2004 The University of Maryland.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. All advertising materials mentioning features or use of this software
     must display the following acknowledgement:
 	This product includes software developed by the University of
 	California, Berkeley and its contributors.
  4. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE. */

#ifndef LIBQUAL_H
#define LIBQUAL_H

/* Call this at any time to reclaim all memory used by libqual and to
   reset libqual to its initial state (i.e., the state when the
   library has just been loaded).  Note: after calling this function,
   you will again need to call init_quals etc. */
void reset_libqual(void);

/* Qualifiers and qualified types */
typedef struct Qtype *qtype;

/* Stores and effects */
typedef struct Abstract_loc *aloc;
typedef struct Effect *effect;
typedef struct Store *store;

/* Polymorphism */
typedef enum {p_neg = -1, p_non = 0, p_pos = 1, p_sub = 2} polarity;
/* jf: p_sub can be used for standard flow */

/* Possible effect constructors */
typedef enum {
  eff_any = 0,
  eff_rwr,
  eff_r,
  eff_wr,
  eff_alloc,
  eff_last = eff_alloc } eff_kind;

/* Other */
typedef int (*printf_func)(const char *fmt, ...);
typedef struct Closure {
  void (*invoke)(struct Closure *);
} *closure;

static inline void invoke_closure(closure c) {
  c->invoke(c);
}

/* Load in a lattice configuration file */
void load_config_file_quals(const char *);

/* jf - shouldn't need stuff past here */
typedef struct Rinfo *rinfo; /* Info for restrict */
typedef struct Drinfo *drinfo; /* Info for deep restrict */
typedef struct ExprDrinfoPair *exprdrinfo; /* Info for deep restrict */

#endif
